/**
*  Ypgrader, Yopy 3000/3500/3700 filesystem and kernel upgrading tool.
*  version 0.2 Sep 11 2003
*  Copyright (C) 2003 Tuomas Tuurala
*
* @author Tuomas Tuurala <tutuural@cc.jyu.fi>
* 
* version 0.2: added --nate option  2003/09/11 Otfried Cheong
*/

/*****************************************************************************
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 *****************************************************************************/

#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>

#include <string.h>
#include <getopt.h>

#include <errno.h>

#include <usb.h>

#define OK 1
#define NOK 0

/* time out in milliseconds  10 minutes*/
#define TIMEOUT (10*60*1000)
const char* program_name;

void print_version(FILE* stream)
{
    fprintf(stream, "Ypgrader, Yopy 3000/3500/3700 filesystem and kernel "
                     "upgrading tool\nVersion 0.2 Sep 12 2003\n");
}

void print_usage(FILE* stream, int exit_code)
{
  print_version(stream);
  fprintf(stream, "Usage: %s options \n", program_name);
  fprintf(stream,
            " -f --filesystem filename  To upgrade filesystem with the file.\n"
            " -k --kernel filename      To upgrade kernel with the file.\n"
	    " -n --nate filename        To upgrade nate fs with the file.\n");
	  
  exit(exit_code);
}

void print_progress(FILE* stream, int count)
{
    fprintf(stream, "\r  %d %%", count);
}

int write_header(usb_dev_handle *yopy, const char *header, int file_size)
{
    int i, size;
    char *tmp = calloc(1, 20 * sizeof(char) );
    if ( tmp == NULL )
    {
        return NOK;
    }

    memcpy(tmp, header, strlen(header) );

    for (i = 0; i < 4; i++ )/*filesize*/
    {
        tmp[19-i] = file_size & 0xFF;
        file_size >>= 8;
    }

    size = usb_bulk_write(yopy, 1, tmp, 20, TIMEOUT);

    free(tmp);

    if ( size != 20 )
    {
        return NOK;
    }
    return OK;
}

int write_image(usb_dev_handle *yopy, FILE *file, int file_size)
{
    int size, total = 0, w;
    char *buf = malloc(1000 * sizeof(char) );
    if ( buf == NULL )
    {
        return NOK;
    }

    while ( 1 )
    {
        size = fread(buf, 1, 1000, file);
        if ( size <=0 )
        {
            break;
        }

        w  = usb_bulk_write(yopy, 1, buf, size, TIMEOUT);
        if ( w != size )
        {
            free(buf);
            return NOK;
        }
        total = total + size;
        print_progress(stdout, (total*100)/file_size);
    }
    fprintf(stdout, "\nDone.\n");
    free(buf);
    return OK;
}

/**
*
* Protocol description:
*     (Protocol Reverse engineering was done by looking the output of
*     usb protocol sniffer when using Windows tool to upgrade Yopy's kernel
*     and filesystem )
*
*     Kernel:
*         1. Upload 20 bytes to Yopy where first 14 bytes contains ascii
*            presentation of header key YPKERNELUPDATE and last 4 bytes contains
*            the image size
*         2. Upload kernel image that is splitted in 1000 bytes blocks
*            to the Yopy
*
*     Filesystem:
*         1. Upload 20 bytes to Yopy where first 15 bytes contains ascii
*            presentation of header key YPFILESYSUPDATE and last 4 bytes contains
*            the image size
*         2. Upload filesystem image that is splitted in 1000 bytes
*            blocks to the Yopy
*/
int upgrade(const char *filename, const char *header, usb_dev_handle *yopy, int wait)
{
    FILE *file;
    struct stat stat_buf;
    int file_size;
    file = fopen(filename, "r");
    if ( file == NULL )
    {
        fprintf(stderr,"%s: %s\n", filename, strerror(errno));
        return NOK;
    }

    fstat( fileno(file), &stat_buf);
    file_size =  stat_buf.st_size;

    fprintf(stdout,"Uploading header key to Yopy.\n");
    if ( wait )
    {
        fprintf(stdout, "Waiting for previous flash update to be completed in Yopy.\n");
    }
    /* first write header key to Yopy */
    if ( write_header(yopy, header, file_size) != OK )
    {
        fprintf(stderr,"Something went wrong when uploading header key to Yopy.\n");
        fprintf(stderr,"%s\n", usb_strerror());
        fprintf(stderr,"Maybe you don't have permissions to write in the usb device\n");
        fclose(file);
        return NOK;
    }

    fprintf(stdout,"Uploading %s image file to Yopy.\n", filename);
    if ( write_image(yopy, file, file_size) != OK )
    {
        fprintf(stderr, "Something went wrong when uploading the image.");
        fprintf(stderr, "%s\n", usb_strerror());
        fclose(file);
        return NOK;
    }

    fclose(file);
    return OK;
}

/**
*  Try to find specific vendor id and product number
*  and open that device
*/
struct usb_dev_handle *open_yopy(void)
{
    struct usb_bus *bus;
    struct usb_device *dev;

    usb_find_busses();
    usb_find_devices();

    for (bus = usb_busses; bus; bus = bus->next)
    {
        for (dev = bus->devices; dev; dev = dev->next)
        {
            if ( dev->descriptor.idVendor == 0xE7E && dev->descriptor.idProduct == 1 )
            {
                return usb_open(dev);
            }
        }
    }
    return NULL;
}


int main(int argc, char* argv[])
{
    int next_option;

    const char* const short_options = "f:k:n:";

    const struct option long_options[] =
    {
        { "kernel",     1, NULL, 'k' },
        { "filesystem", 1, NULL, 'f' },
	{ "nate",       1, NULL, 'n' },
        { NULL,         0, NULL,  0  }
    };

    const char *kernel_filename = NULL;
    const char *filesystem_filename = NULL;
    const char *nate_filename = NULL;
    usb_dev_handle *yopy;
    program_name = argv[0];
    int num_updates = 0;
    int wait = 0;
    do
    {
        next_option =
            getopt_long(argc, argv, short_options, long_options, NULL);
        switch (next_option)
        {
            case 'f':
                filesystem_filename = optarg;
                break;
            case 'k':
                kernel_filename = optarg;
                break;
            case 'n':
                nate_filename = optarg;
                break;
            case -1:
                break;
            default:
                print_usage(stderr, 1);
        }
    }
    while (next_option != -1);

    if ( filesystem_filename == NULL && kernel_filename == NULL 
	 && nate_filename == NULL)
    {
        print_usage(stderr, 1);
    }

    print_version(stdout);

    if ( kernel_filename )
      num_updates += 1;
    if ( filesystem_filename )
      num_updates += 1;
    if ( nate_filename )
      num_updates += 1;
    
    wait = (num_updates > 1) ? 1 : 0;

    usb_init();

    fprintf(stdout, "Try to open connection to Yopy.\n");

    yopy = open_yopy();

    if ( yopy == NULL )
    {
        fprintf(stderr, "Cannot find Yopy.\n");
        return 1;
    }
    fprintf(stdout, "Yopy found.\n");

    /* Upgrade kernel */
    if ( kernel_filename )
    {
        fprintf(stdout, "Upgrading kernel.\n");
        if ( upgrade(kernel_filename, "YPKERNELUPDATE", yopy, 0) == NOK )
        {
            return 1;
        }
    }

    /* Upgrade filesystem */
    if ( filesystem_filename )
    {
        fprintf(stdout, "Upgrading filesystem.\n");
        if ( upgrade(filesystem_filename, "YPFILESYSUPDATE", yopy, wait) == NOK )
        {
            return 1;
        }
    }

    /* Upgrade nate filesystem */
    if ( nate_filename )
    {
        fprintf(stdout, "Upgrading nate (mtdblock3) filesystem.\n");
        if ( upgrade(nate_filename, "YPNATEUPDATE", yopy, wait) == NOK )
        {
            return 1;
        }
    }

    usb_close(yopy);

    return 0;
}


